try:
    from functools import cached_property
except ImportError:
    from threading import RLock

    _NOT_FOUND = object()

    class cached_property:
        def __init__(self, func):
            self.func = func
            self.attrname = None
            self.__doc__ = func.__doc__
            self.lock = RLock()

        def __set_name__(self, owner, name):
            if self.attrname is None:
                self.attrname = name
            elif name != self.attrname:
                raise TypeError('Cannot assign the same cached_property to two different names ' f'({self.attrname!r} and {name!r}).')

        def __get__(self, instance, owner=None):
            if instance is None:
                return self
            if self.attrname is None:
                raise TypeError('Cannot use cached_property instance without calling __set_name__ on it.')
            try:
                cache = instance.__dict__
            except AttributeError:  # not all objects have __dict__ (e.g. class defines slots)
                msg = f"No '__dict__' attribute on {type(instance).__name__!r} " f'instance to cache {self.attrname!r} property.'
                raise TypeError(msg) from None
            val = cache.get(self.attrname, _NOT_FOUND)
            if val is _NOT_FOUND:
                with self.lock:
                    # check if another thread filled cache while we awaited lock
                    val = cache.get(self.attrname, _NOT_FOUND)
                    if val is _NOT_FOUND:
                        val = self.func(instance)
                        try:
                            cache[self.attrname] = val
                        except TypeError:
                            msg = (
                                f"The '__dict__' attribute on {type(instance).__name__!r} instance "
                                f'does not support item assignment for caching {self.attrname!r} property.'
                            )
                            raise TypeError(msg) from None
            return val



# Generated by Django 2.2.28 on 2022-04-12 13:16
import redis
from django.db import migrations


def forwards_func(apps, schema_editor):
    RedisServer = apps.get_model('redisboard', 'RedisServer')
    db_alias = schema_editor.connection.alias
    for obj in RedisServer.objects.using(db_alias).all():
        if obj.hostname.startswith('/'):
            obj.url = f'unix://{obj.hostname}'
        else:
            obj.url = f'redis://{obj.hostname}:{obj.port}'
        obj.save(update_fields=['url'])


def reverse_func(apps, schema_editor):
    RedisServer = apps.get_model('redisboard', 'RedisServer')
    db_alias = schema_editor.connection.alias
    for obj in RedisServer.objects.using(db_alias).all():
        kwargs = redis.connection.parse_url(obj.url)
        if kwargs['connection_class'] is redis.UnixDomainSocketConnection:
            obj.hostname = kwargs['path']
            obj.save(update_fields=['hostname'])
        else:
            obj.hostname = kwargs['hostname']
            obj.port = kwargs['port']
            obj.save(update_fields=['hostname', 'port'])


class Migration(migrations.Migration):
    dependencies = [
        ('redisboard', '0002_add_url'),
    ]

    operations = [
        migrations.RunPython(forwards_func, reverse_func),
    ]



# Generated by Django 3.2.13 on 2022-05-15 15:29

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('redisboard', '0004_cleanup'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='redisserver',
            name='sampling_size',
        ),
        migrations.RemoveField(
            model_name='redisserver',
            name='sampling_threshold',
        ),
    ]



# -*- coding: utf-8 -*-


import django.core.validators
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='RedisServer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label', models.CharField(blank=True, max_length=50, verbose_name='Label', null=True)),
                (
                    'hostname',
                    models.CharField(
                        verbose_name='Hostname', max_length=250, help_text='This can also be the absolute path to a redis socket'
                    ),
                ),
                (
                    'port',
                    models.IntegerField(
                        default=6379,
                        validators=[django.core.validators.MaxValueValidator(65535), django.core.validators.MinValueValidator(1)],
                        verbose_name='Port',
                        blank=True,
                        null=True,
                    ),
                ),
                ('password', models.CharField(blank=True, max_length=250, verbose_name='Password', null=True)),
                (
                    'sampling_threshold',
                    models.IntegerField(
                        default=1000,
                        verbose_name='Sampling threshold',
                        help_text='Number of keys after which only a sample (of random keys) is shown on the inspect page.',
                    ),
                ),
                (
                    'sampling_size',
                    models.IntegerField(
                        default=200,
                        verbose_name='Sampling size',
                        help_text='Number of random keys shown when sampling is used. Note that each key translates to a RANDOMKEY call'
                        ' in redis.',
                    ),
                ),
            ],
            options={
                'verbose_name_plural': 'Redis Servers',
                'verbose_name': 'Redis Server',
                'permissions': (('can_inspect', 'Can inspect redis servers'),),
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='redisserver',
            unique_together=set([('hostname', 'port')]),
        ),
    ]



from datetime import datetime
from datetime import timedelta
from typing import Dict
from typing import NamedTuple
from typing import Union

from attr import define
from attr import field
from attr.converters import optional


def timedelta_fromseconds(sec):
    return timedelta(seconds=sec)


def int_if_positive(val):
    val = int(val)
    if val >= 0:
        return val


def dash_if_none(val):
    if val is None:
        return '-'
    else:
        return val


def ascii_if_not_none(val):
    if val is not None:
        return val.decode()


def timedelta_fromseconds_if_positive(sec):
    if sec >= 0:
        return timedelta(seconds=sec)


def datetime_fromtimestamp_usec(time_usec):
    return datetime.fromtimestamp(time_usec / 1000_000_000)


@define
class KeyInfo:
    name: str = field()
    type: str = field()
    encoding: str = field()
    ttl: timedelta = field(converter=timedelta_fromseconds_if_positive)
    length: int = field(converter=int_if_positive)
    frequency: Union[int, None] = field(converter=optional(int), default=None)
    idletime: Union[timedelta, None] = field(converter=optional(timedelta_fromseconds), default=None)


class ScanResult(NamedTuple):
    cursor: int  # next cursor
    count: int
    total: int
    data: object


@define
class DBInfo:
    id: int
    stats: Dict[str, Union[str, int]]
    cursor: int = None  # current cursor
    scan: Union[bool, ScanResult] = False



import pytest
from process_tests import TestProcess
from process_tests import wait_for_strings
from redis.client import StrictRedis


@pytest.fixture
def redis_server(tmp_path):
    redis_socket = str(tmp_path.joinpath('redis.sock'))
    with TestProcess(
        'redis-server', '--port', '0', '--save', '', '--appendonly', 'yes', '--dir', tmp_path, '--unixsocket', redis_socket
    ) as redis_server:
        wait_for_strings(redis_server.read, 2, 'ready to accept connections')
        yield redis_socket
        print(redis_server.read())


@pytest.fixture
def redis_monitor(redis_server):
    with TestProcess('redis-cli', '-s', redis_server, 'monitor') as redis_monitor:
        yield
        print(redis_monitor.read())


@pytest.fixture
def redis_conn(redis_server):
    return StrictRedis(unix_socket_path=redis_server, decode_responses=True)



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mredisboard` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``redisboard.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``redisboard.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""

import argparse
import os
import random
import string

from django.contrib.admin import AdminSite
from django.contrib.admin.apps import AdminConfig

from . import __version__


def parse_class(value):
    if '.' not in value:
        value = f'redisboard.data.{value}'
    return value


parser = argparse.ArgumentParser(description='Runs redisboard in an ad-hoc django project.')
parser.add_argument(
    '--debug',
    action='store_true',
    help='Enable debug mode (DEBUG=True and autoreload enabled).',
)
parser.add_argument(
    '--password',
    '-p',
    help='The admin password. A random one will be generated if not provided.',
)
parser.add_argument(
    '--storage',
    '-s',
    default=os.path.expanduser('~/.redisboard'),
    help='Where to save the SECRET_KEY and sqlite database. (default: %(default)s)',
)
parser.add_argument(
    '--decoder',
    '-d',
    default='UTF8BackslashReplaceDecoder',
    type=parse_class,
    help='Decoder class to use . (default: %(default)s)',
)
parser.add_argument(
    '--version',
    action='version',
    version=f'redisboard {__version__}',
)
parser.add_argument(
    'addrport',
    nargs='?',
    default='0:8000',
    help='Optional port number, or ipaddr:port (default: %(default)s)',
)

DJANGO_SETTINGS = dict(
    REDISBOARD_SOCKET_CONNECT_TIMEOUT=5,
    REDISBOARD_SOCKET_TIMEOUT=5,
    CACHES={
        'default': {
            'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        },
    },
    SESSION_ENGINE='django.contrib.sessions.backends.cached_db',
    INSTALLED_APPS=(
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'django.contrib.sessions',
        'django.contrib.messages',
        'django.contrib.staticfiles',
        'redisboard.cli.RedisboardAdminConfig',
        'redisboard',
    ),
    MIDDLEWARE=[
        'django.middleware.security.SecurityMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.common.CommonMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
    ],
    ROOT_URLCONF=__name__,
    STATIC_URL='/static/',
    ALLOWED_HOSTS=['*'],
    LOGGING={
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'default': {
                'format': '[%(asctime)s] %(levelname)s | %(name)s - %(message)s',
                'datefmt': '%d/%b/%Y %H:%M:%S',
            }
        },
        'handlers': {
            'console': {'level': 'DEBUG', 'class': 'logging.StreamHandler', 'formatter': 'default'},
        },
        'loggers': {
            'django.request': {
                'handlers': ['console'],
                'level': 'DEBUG',
                'propagate': True,
            },
            '': {'handlers': ['console'], 'level': 'INFO', 'propagate': False},
        },
    },
)


class RedisboardAdminSite(AdminSite):
    site_header = f'Redisboard v{__version__}'


class RedisboardAdminConfig(AdminConfig):
    default_site = 'redisboard.cli.RedisboardAdminSite'


def get_random(length=50, chars='abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)'):
    return ''.join(random.SystemRandom().choice(chars) for _ in range(length))


def main(args=None):
    args = parser.parse_args(args=args)
    if not os.path.exists(args.storage):
        os.mkdir(args.storage)
    secret_key_path = os.path.join(args.storage, 'SECRET_KEY')
    if os.path.exists(secret_key_path):
        with open(secret_key_path) as fh:
            secret_key = fh.read()
    else:
        secret_key = get_random()
        with open(secret_key_path, 'w') as fh:
            fh.write(secret_key)
    database_path = os.path.join(args.storage, 'db.sqlite')

    import django
    from django.conf import settings
    from django.contrib import admin
    from django.contrib.staticfiles.storage import staticfiles_storage
    from django.core.management import execute_from_command_line
    from django.urls import path
    from django.views.generic.base import RedirectView

    settings.configure(
        DATABASES={'default': {'ENGINE': 'django.db.backends.sqlite3', 'NAME': database_path}},
        DEBUG=args.debug,
        REDISBOARD_DECODER_CLASS=args.decoder,
        SECRET_KEY=secret_key,
        TEMPLATES=[
            {
                'BACKEND': 'django.template.backends.django.DjangoTemplates',
                'DIRS': [],
                'APP_DIRS': True,
                'OPTIONS': {
                    'debug': args.debug,
                    'string_if_invalid': '[MISSING VARIABLE: "%s"]' if args.debug else '',
                    'context_processors': [
                        'django.template.context_processors.debug',
                        'django.template.context_processors.request',
                        'django.contrib.auth.context_processors.auth',
                        'django.contrib.messages.context_processors.messages',
                    ],
                },
            }
        ],
        **DJANGO_SETTINGS,
    )

    django.setup()

    global urlpatterns
    urlpatterns = [
        path('favicon.ico', RedirectView.as_view(url=staticfiles_storage.url('redisboard/favicon.ico'), permanent=True)),
        path('', admin.site.urls),
    ]

    from django.contrib.auth.models import User

    from redisboard.models import RedisServer

    create_user = not os.path.exists(database_path)
    execute_from_command_line(['django-admin', 'migrate', '--noinput'])
    if create_user:
        user = User(username='redisboard', is_superuser=True, is_staff=True, is_active=True)
        pwd = get_random(8, string.digits + string.ascii_letters) if args.password is None else args.password
        user.set_password(pwd)
        user.save()
        print()
        print('=' * 80)
        print(
            f"""
    Credentials:

        USERNAME: redisboard
        PASSWORD: {pwd if args.password is None else '<PROVIDED VIA --password OPTION>'}

"""
        )
        print('=' * 80)
        print()

        RedisServer.objects.create(label='localhost', url='redis://127.0.0.1')
    elif args.password is not None:
        user = User.objects.get(username='redisboard')
        if not user.check_password(args.password):
            print('Updating password...')
            user.set_password(args.password)
            user.save(update_fields=['password'])

    if args.debug:
        execute_from_command_line(['django-admin', 'runserver', '--insecure', args.addrport])
    else:
        execute_from_command_line(['django-admin', 'runserver', '--insecure', '--noreload', args.addrport])



import re
from datetime import datetime
from itertools import starmap
from logging import getLogger
from typing import TYPE_CHECKING
from typing import Dict
from typing import Type

import redis
from attr import Factory
from attr import define
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from .connection import ClosableStrictRedis
from .structs import datetime_fromtimestamp_usec
from .structs import timedelta_fromseconds
from .utils import cached_property

if TYPE_CHECKING:
    from .data import BaseDecoder
    from .data import BaseDisplay
    from .data import LengthQuery
    from .data import ValueQuery

logger = getLogger(__name__)

REDISBOARD_DECODER_CLASS: 'Type[BaseDecoder]' = import_string(
    getattr(settings, 'REDISBOARD_DECODER_CLASS', 'redisboard.data.UTF8BackslashReplaceDecoder'),
)
REDISBOARD_DISPLAY_CLASS: 'Type[BaseDisplay]' = import_string(
    getattr(settings, 'REDISBOARD_DISPLAY_CLASS', 'redisboard.data.TabularDisplay'),
)
REDISBOARD_VALUE_QUERY_CLASS: 'Type[ValueQuery]' = import_string(
    getattr(settings, 'REDISBOARD_VALUE_QUERY_CLASS', 'redisboard.data.ValueQuery'),
)
REDISBOARD_LENGTH_QUERY_CLASS: 'Type[LengthQuery]' = import_string(
    getattr(settings, 'REDISBOARD_LENGTH_QUERY_CLASS', 'redisboard.data.LengthQuery'),
)

REDISBOARD_DETAIL_FILTERS = [
    re.compile(name)
    for name in getattr(
        settings,
        'REDISBOARD_DETAIL_FILTERS',
        (
            '.*_version',
            '.*_api',
            '.*_human',
            '.*_policy',
            'expired_keys',
            'evicted_keys',
            'redis_version',
            'redis_mode',
            'os',
        ),
    )
]
REDISBOARD_DETAIL_CONVERTERS = {
    re.compile(name): converter
    for name, converter in getattr(
        settings,
        'REDISBOARD_DETAIL_CONVERTERS',
        {
            '.*_seconds$': timedelta_fromseconds,
            'avg_ttl': timedelta_fromseconds,
            '.*_time$': datetime.fromtimestamp,
            '.*_time_usec$': datetime_fromtimestamp_usec,
        },
    ).items()
}

REDISBOARD_SLOWLOG_NUM = getattr(settings, 'REDISBOARD_SLOWLOG_NUM', 10)


def coerce_detail(key, value):
    for name, converter in REDISBOARD_DETAIL_CONVERTERS.items():
        if name.match(key):
            try:
                return key, converter(value)
            except Exception:
                logger.exception(f'Failed converting {key}={value} with {converter}')
    else:
        return key, value


def validate_url(value):
    try:
        redis.connection.parse_url(value)
    except Exception as exc:
        raise ValidationError(str(exc))


@define(slots=False)
class RedisServerStats:
    status: str = 'n/a'
    info: dict = Factory(dict)
    slowlog: list = Factory(dict)

    @cached_property
    def details(self):
        return dict(coerce_detail(k, v) for k, v in self.info.items() if any(name.match(k) for name in REDISBOARD_DETAIL_FILTERS))

    @cached_property
    def memory(self):
        if 'used_memory_human' in self.info:
            return f'{self.info["used_memory_human"]} (peak: {self.info.get("used_memory_peak_human", "n/a")})'
        else:
            return 'n/a'

    @cached_property
    def clients(self):
        return self.info.get('connected_clients', 'n/a')

    @cached_property
    def databases(self) -> Dict[int, dict]:
        return {int(name[2:]): dict(starmap(coerce_detail, data.items())) for name, data in self.info.items() if name.startswith('db')}

    def __bool__(self):
        return self.status == 'UP'


class RedisServer(models.Model):
    class Meta:
        verbose_name = _('Redis Server')
        verbose_name_plural = _('Redis Servers')
        permissions = (('can_inspect', 'Can inspect redis servers'),)

    label = models.CharField(
        verbose_name=_('Label'),
        max_length=50,
        blank=True,
        null=True,
    )
    url = models.CharField(
        verbose_name=_('URL'),
        max_length=250,
        unique=True,
        help_text=_(
            '<a href="https://www.iana.org/assignments/uri-schemes/prov/redis">IANA-compliant</a> URL. Examples: <pre>'
            'redis://[[username]:[password]]@localhost:6379/0\n'
            'rediss://[[username]:[password]]@localhost:6379/0\n'
            'unix://[[username]:[password]]@/path/to/socket.sock?db=0</pre>'
        ),
        validators=[validate_url],
    )
    password = models.CharField(
        verbose_name=_('Password'),
        max_length=250,
        null=True,
        blank=True,
        help_text=_('You can also specify the password here (the field is masked).'),
    )

    @cached_property
    def connection(self) -> redis.StrictRedis:
        return ClosableStrictRedis(self.url, self.password)

    @cached_property
    def display(self) -> 'BaseDisplay':
        return REDISBOARD_DISPLAY_CLASS(
            decoder_class=REDISBOARD_DECODER_CLASS,
            value_query_class=REDISBOARD_VALUE_QUERY_CLASS,
            length_query_class=REDISBOARD_LENGTH_QUERY_CLASS,
            server=self,
        )

    @cached_property
    def stats(self) -> RedisServerStats:
        try:
            conn = self.connection
            info = conn.info()
            slowlog = conn.slowlog_get(num=REDISBOARD_SLOWLOG_NUM)
            return RedisServerStats(
                status='UP',
                info=info,
                slowlog=slowlog,
            )
        except (redis.exceptions.ConnectionError, redis.exceptions.TimeoutError) as exc:
            return RedisServerStats(
                status=f'DOWN: {exc}',
            )
        except redis.exceptions.RedisError as exc:
            return RedisServerStats(
                status=f'ERROR: {exc!r}',
            )

    @cached_property
    def has_frequency(self):
        return self.stats.info['maxmemory_policy'].endswith('-lfu')

    def __str__(self):
        if self.label:
            return self.label
        else:
            return self.url



import html
import pickle
from abc import ABC
from abc import abstractmethod
from functools import partial
from itertools import chain
from typing import TYPE_CHECKING
from typing import Dict
from typing import List
from typing import Tuple
from typing import Type
from urllib.parse import quote

from django.conf import settings
from django.shortcuts import resolve_url
from django.template.defaultfilters import truncatechars
from django.utils.safestring import mark_safe
from django.utils.translation import gettext
from redis import StrictRedis

from redisboard.structs import KeyInfo
from redisboard.structs import ScanResult
from redisboard.structs import ascii_if_not_none
from redisboard.structs import dash_if_none

if TYPE_CHECKING:
    from redisboard.models import RedisServer

REDISBOARD_SCAN_COUNT: int = getattr(settings, 'REDISBOARD_SCAN_COUNT', 1000)
REDISBOARD_STRING_PAGINATION: int = getattr(settings, 'REDISBOARD_STRING_PAGINATION', 10000)


class BaseDecoder(ABC):
    def __init__(self, server):
        pass

    def key(self, key: bytes):
        return key.decode()

    def bytes(self, key: str, value: bytes):
        return value

    def string(self, key: str, value: bytes, **kwargs):
        return [(len(value), self.bytes(key, value))]

    def hash(self, key: str, hash_value: Dict[bytes, bytes], **kwargs):
        return sorted((self.hash_field(key, k), self.bytes(key, v)) for k, v in hash_value.items())

    def hash_field(self, key: str, field: bytes):
        return self.key(field)

    def list(self, key: str, value: List[bytes], count=int, **kwargs):
        return list(enumerate((self.bytes(key, v) for v in value), start=count))

    def set(self, key: str, value: List[bytes], count: int, **kwargs):
        return list(enumerate(sorted(self.bytes(key, v) for v in value), start=count))

    def zset(self, key: str, value: bytes, **kwargs):
        return sorted((s, self.bytes(key, v)) for v, s in value)

    def unsupported(self, key: str, value, type_: str, **kwargs):
        return [('ERROR', f'Missing decoder; {value}')]

    def __getattr__(self, type_: str):
        return partial(self.unsupported, type_=type_)


class UTF8BackslashReplaceDecoder(BaseDecoder):
    def bytes(self, key: str, value: bytes):
        return value.decode(errors='backslashreplace')

    def key(self, key: bytes):
        return key.decode(errors='backslashreplace')


class PickleDecoder(BaseDecoder):
    def bytes(self, key: str, value: bytes):
        return pickle.loads(value)


class ValueQuery:
    connection: StrictRedis

    def __init__(self, connection: StrictRedis):
        self.connection = connection

    def hash(self, key, *, cursor=0, **kwargs) -> Tuple[int, list]:
        return self.connection.hscan(key, cursor=cursor, count=REDISBOARD_SCAN_COUNT)

    def list(self, key, *, cursor=0, **kwargs) -> Tuple[int, list]:
        end = cursor + REDISBOARD_SCAN_COUNT
        value = self.connection.lrange(key, cursor, end - 1)
        if len(value) < end:
            return 0, value
        else:
            return end, value

    def set(self, key, *, cursor=0, **kwargs) -> Tuple[int, list]:
        return self.connection.sscan(key, cursor=cursor, count=REDISBOARD_SCAN_COUNT)

    def string(self, key, *, cursor=0, count=0, **kwargs) -> Tuple[int, list]:
        start = cursor
        end = count + REDISBOARD_STRING_PAGINATION
        value = self.connection.getrange(key, start, end - 1)
        if value:
            return end, value
        else:
            return 0, value

    def zset(self, key, *, cursor=0, **kwargs) -> Tuple[int, list]:
        return self.connection.zscan(key, cursor=cursor, count=REDISBOARD_SCAN_COUNT)

    def unsupported(self, key, *, cursor=0, type_, **kwargs):
        return cursor, f'Unsupported type {type_!r} for key: {key}'

    def __getattr__(self, type_):
        return partial(self.unsupported, type_=type_)


class LengthQuery:
    connection: StrictRedis

    def __init__(self, connection):
        self.connection = connection

    def string(self, key):
        return self.connection.strlen(key)

    def hash(self, key):
        return self.connection.hlen(key)

    def list(self, key):
        return self.connection.llen(key)

    def set(self, key):
        return self.connection.scard(key)

    def zset(self, key):
        return self.connection.zcount(key, '-inf', '+inf')

    def unsupported(self, key):
        return self.connection.echo('-1')

    def __getattr__(self, type_):
        return self.unsupported


class BaseDisplay(ABC):
    decoder: BaseDecoder
    value_query_class: Type[ValueQuery]
    length_query_class: Type[LengthQuery]
    server: 'RedisServer'

    def __init__(
        self,
        decoder_class: Type[BaseDecoder],
        value_query_class: Type[ValueQuery],
        length_query_class: Type[LengthQuery],
        server: 'RedisServer',
    ):
        self.decoder = decoder_class(server)
        self.value_query_class = value_query_class
        self.length_query_class = length_query_class
        self.server = server

    @abstractmethod
    def slowlog(self):
        pass

    @abstractmethod
    def details(self):
        pass

    @abstractmethod
    def cpu(self):
        pass

    def keys(self, db, keys):
        conn = self.server.connection
        conn.select(db)

        if self.server.has_frequency:
            usage_field, usage_command = 'frequency', 'FREQ'
        else:
            usage_field, usage_command = 'idletime', 'IDLETIME'

        with conn.pipeline() as pipe:
            for key in keys:
                pipe.type(key)
                pipe.object('ENCODING', key)
            result = iter(map(ascii_if_not_none, pipe.execute()))
            values = list(zip(keys, result, result))

        with conn.pipeline() as pipe:
            query = self.length_query_class(pipe)
            for key, type_, *_ in values:
                pipe.ttl(key)
                pipe.object(usage_command, key)
                getattr(query, type_)(key)
            result = iter(pipe.execute())

        fields = ['name', 'type', 'encoding', 'ttl', usage_field, 'length']
        values = list(map(list, map(chain.from_iterable, zip(values, zip(result, result, result)))))

        return [KeyInfo(**dict(zip(fields, v))) for v in values]

    def scan(self, db, cursor=0, match=None, type=None) -> ScanResult:
        conn = self.server.connection
        conn.select(db)

        total = conn.dbsize()
        cursor, keys = conn.scan(cursor=cursor, count=REDISBOARD_SCAN_COUNT, match=match, _type=type)
        return ScanResult(cursor, len(keys), total, self.keys(db, keys))

    def value(self, db, key, **kwargs):
        conn = self.server.connection
        conn.select(db)
        type_ = conn.type(key).decode()
        total = getattr(self.length_query_class(conn), type_)(key)
        cursor, value = getattr(self.value_query_class(conn), type_)(key, **kwargs)
        decoded_value = getattr(self.decoder, type_)(key, value, **kwargs)
        return ScanResult(cursor, len(value), total, decoded_value)


class TabularDisplay(BaseDisplay):
    def slowlog(self):
        commands = []
        slowlog = self.server.stats.slowlog
        for log in slowlog:
            command = log['command']
            if isinstance(command, bytes):
                command = command.decode(errors='backslashreplace')

            if len(command) > 255:
                command = f'{command:252}...'

            commands.append((log['duration'], command))
        commands.sort(reverse=True)
        if commands:
            output = ''.join(f'<tr><th>{duration / 1000.0:.1f}ms</th><td>{command}</td></tr>' for duration, command in commands)
            return mark_safe(f'<table><tr><th colspan="2">Total: {len(slowlog)} items</th></tr>{output}</table>')
        else:
            return 'n/a'

    def details(self):
        stats = self.server.stats
        details = stats.details
        output = [f'<tr><th>{k.replace("_", " ")}</th><td>{v}</td></tr>' for k, v in details.items()]
        for db, details in stats.databases.items():
            keys = details.keys()
            output.append(f'<tr><td colspan="2"><table><tr><th>db{db}</th>')
            output.extend(f'<th>{k}</th>' for k in keys)
            output.append('</tr><tr><td></td>')
            output.extend(f'<td>{details[k]}</td>' for k in keys)
            output.append('</tr></table></td></tr>')
        if output:
            return mark_safe(f'<table>{"".join(output)}</table>')
        return 'n/a'

    def cpu(self):
        stats = self.server.stats
        if stats.status != 'UP':
            return 'n/a'

        data = (
            'used_cpu_sys',
            'used_cpu_sys_children',
            'used_cpu_user',
            'used_cpu_user_children',
        )
        data = dict((k, stats.info[k]) for k in data)
        total_cpu = sum(data.values())
        uptime = stats.info['uptime_in_seconds']
        data['cpu_utilization'] = f'{total_cpu / uptime if uptime else 0:.3f}%%'

        data = sorted(data.items())

        output = []
        for k, v in data:
            k = k.replace('_', ' ')
            output.append(f'<tr><th>{k}</th><td>{v}</td></tr>')

        if output:
            return mark_safe(f'<table>{"".join(output)}</table>')
        return 'n/a'

    def keys(self, db, *args):
        keys = super().keys(db, *args)
        has_frequency = self.server.has_frequency
        output = [
            '<table><thead><tr><th rowspan="2">',
            gettext('Keys'),
            f' ({len(keys)})</th><th colspan="7">',
            gettext('Details'),
            '</th></tr><tr><th>',
            gettext('Type'),
            '</th><th>',
            gettext('TTL'),
            '</th><th>',
            gettext('Encoding'),
            '</th><th>',
            gettext('Length'),
            '</th><th>',
            gettext('Frequency') if self.server.has_frequency else gettext('Idletime'),
            '</th></tr></thead>',
        ]

        output.extend(
            f'<tr><td><a href="'
            f'{resolve_url("admin:redisboard_redisserver_inspect", server_id=self.server.id, db=db, key=quote(key.name))}">'
            f'{html.escape(truncatechars(self.decoder.key(key.name), 200))}</a></td>'
            f'<td>{key.type}</td>'
            f'<td>{dash_if_none(key.ttl)}</td>'
            f'<td>{dash_if_none(key.encoding)}</td>'
            f'<td>{dash_if_none(key.length)}</td>'
            f'<td>{key.frequency if has_frequency else key.idletime}</td>'
            '</tr>'
            for key in keys
        )
        output.append('</table>')
        return mark_safe(''.join(output))

    def value(self, db, key, **kwargs):
        cursor, count, total, data = super().value(db, key, **kwargs)
        return ScanResult(
            cursor,
            count,
            total,
            mark_safe(
                ''.join(
                    chain(
                        ('<table>',),
                        (f'<tr><th>{html.escape(str(k))}</th><td>{html.escape(str(v))}' for k, v in data),
                        ('</table>',),
                    )
                )
            ),
        )



import os

import django
import redis

if __name__ == '__main__':
    os.putenv('DJANGO_SETTINGS_MODULE', 'test_project.settings')
    django.setup()
    conn = redis.StrictRedis()
    # conn.flushall()
    # conn.select(9)
    # conn.flushdb()
    import test_redisboard

    conn.set('monotonic:str', b'\n'.join(f'item-{i}'.encode() for i in range(10000)))
    test_redisboard.make_test_data(conn, size=15000)
    conn.select(1)
    for i in range(15000):
        conn.set(i, i)

    conn.eval(
        """
        local c = 0
        while c < 1000000 do
            redis.call('get', 'str')
            c = c + 1
        end
        """,
        0,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import subprocess
import sys
from os.path import abspath
from os.path import dirname
from os.path import exists
from os.path import join
from os.path import relpath

base_path = dirname(dirname(abspath(__file__)))
templates_path = join(base_path, 'ci', 'templates')


def check_call(args):
    print('+', *args)
    subprocess.check_call(args)


def exec_in_env():
    env_path = join(base_path, '.tox', 'bootstrap')
    if sys.platform == 'win32':
        bin_path = join(env_path, 'Scripts')
    else:
        bin_path = join(env_path, 'bin')
    if not exists(env_path):
        import subprocess

        print('Making bootstrap env in: {0} ...'.format(env_path))
        try:
            check_call([sys.executable, '-m', 'venv', env_path])
        except subprocess.CalledProcessError:
            try:
                check_call([sys.executable, '-m', 'virtualenv', env_path])
            except subprocess.CalledProcessError:
                check_call(['virtualenv', env_path])
        print('Installing `jinja2` into bootstrap environment...')
        check_call([join(bin_path, 'pip'), 'install', 'jinja2', 'tox'])
    python_executable = join(bin_path, 'python')
    if not os.path.exists(python_executable):
        python_executable += '.exe'

    print('Re-executing with: {0}'.format(python_executable))
    print('+ exec', python_executable, __file__, '--no-env')
    os.execv(python_executable, [python_executable, __file__, '--no-env'])


def main():
    import jinja2

    print('Project path: {0}'.format(base_path))

    jinja = jinja2.Environment(
        loader=jinja2.FileSystemLoader(templates_path),
        trim_blocks=True,
        lstrip_blocks=True,
        keep_trailing_newline=True,
    )

    tox_environments = [
        line.strip()
        # 'tox' need not be installed globally, but must be importable
        # by the Python that is running this script.
        # This uses sys.executable the same way that the call in
        # cookiecutter-pylibrary/hooks/post_gen_project.py
        # invokes this bootstrap.py itself.
        for line in subprocess.check_output([sys.executable, '-m', 'tox', '--listenvs'], universal_newlines=True).splitlines()
    ]
    tox_environments = [line for line in tox_environments if line.startswith('py')]

    for root, _, files in os.walk(templates_path):
        for name in files:
            relative = relpath(root, templates_path)
            with open(join(base_path, relative, name), 'w') as fh:
                fh.write(jinja.get_template(join(relative, name)).render(tox_environments=tox_environments))
            print('Wrote {}'.format(name))
    print('DONE.')


if __name__ == '__main__':
    args = sys.argv[1:]
    if args == ['--no-env']:
        main()
    elif not args:
        exec_in_env()
    else:
        print('Unexpected arguments {0}'.format(args), file=sys.stderr)
        sys.exit(1)



import warnings
from logging import getLogger
from traceback import format_stack

from django.conf import settings
from redis import ConnectionPool
from redis import StrictRedis

logger = getLogger(__name__)

REDISBOARD_CONNECTION_POOL_OPTIONS = getattr(settings, 'REDISBOARD_CONNECTIONPOOL_OPTIONS', {})


class ClosableStrictRedis(StrictRedis):
    connection_pool: ConnectionPool

    def __init__(self, url, password):
        super().__init__(
            single_connection_client=True,
            connection_pool=ConnectionPool.from_url(
                url,
                password=password,
                **REDISBOARD_CONNECTION_POOL_OPTIONS,
            ),
        )
        self.created_from = ''.join(format_stack(limit=50))

    def __del__(self):
        if getattr(self, 'connection', None) is not None and settings.DEBUG:
            warnings.warn(
                f'Redis connection was closed from __del__!\n\n'
                f'Created from:\n{self.created_from}\n'
                f'Current stack:\n{"".join(format_stack(limit=25))}',
                ResourceWarning,
                source=self,
            )
        self.close()

    def __exit__(self, *args):
        if not hasattr(self, 'connection'):
            return

        conn = self.connection
        if conn:
            conn.disconnect()
            self.connection = None
            self.connection_pool.release(conn)
        self.connection_pool.disconnect()

    close = __exit__



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(join(dirname(__file__), *names), encoding=kwargs.get('encoding', 'utf8')) as fh:
        return fh.read()


setup(
    name='django-redisboard',
    version='8.2.0',
    license='BSD-2-Clause',
    description='Redis monitoring and inspection drop-in application using django admin.',
    long_description='{}\n{}'.format(
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst')),
    ),
    author='Ionel Cristian Mărieș',
    author_email='contact@ionelmc.ro',
    url='https://github.com/ionelmc/django-redisboard',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        'Topic :: Utilities',
        'Framework :: Django',
        'Topic :: Internet :: WWW/HTTP',
    ],
    project_urls={
        'Documentation': 'https://django-redisboard.readthedocs.io/',
        'Changelog': 'https://django-redisboard.readthedocs.io/en/latest/changelog.html',
        'Issue Tracker': 'https://github.com/ionelmc/django-redisboard/issues',
    },
    keywords=['django', 'redis', 'monitoring', 'inspector', 'statistics'],
    python_requires='>=3.7',
    install_requires=[
        'redis>=3.5',
        'Django>=2.2',
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    entry_points={
        'console_scripts': [
            'redisboard = redisboard.cli:main',
        ]
    },
)



from functools import partial
from functools import wraps
from logging import getLogger
from typing import Union
from urllib.parse import quote
from urllib.parse import unquote_to_bytes

import redis
from django.contrib import admin
from django.http import HttpResponseForbidden
from django.http import HttpResponseNotFound
from django.shortcuts import get_object_or_404
from django.shortcuts import render
from django.shortcuts import resolve_url
from django.template.response import TemplateResponse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from .data import REDISBOARD_SCAN_COUNT
from .models import RedisServer
from .structs import DBInfo

logger = getLogger(__name__)


def cleanup_changelist_response(response: TemplateResponse):
    obj: RedisServer
    for obj in response.context_data['cl'].queryset:
        cleanup_connection(response, obj)


def cleanup_connection(_, server: RedisServer):
    connection = server.__dict__.get('connection')
    if connection:
        connection.close()


class RedisServerAdmin(admin.ModelAdmin):
    class Media:
        css = {'all': ('redisboard/admin.css',)}

    list_display = (
        'display_name',
        'status',
        'memory',
        'clients',
        'details',
        'cpu_utilization',
        'slowlog',
        'tools',
    )

    list_filter = ('label',)

    def display_name(self, obj: RedisServer):
        return str(obj)

    display_name.ordering = 'label', 'url'
    display_name.short_description = _('Name')

    def details(self, obj: RedisServer):
        return obj.display.details()

    details.short_description = _('Details')

    def cpu_utilization(self, obj: RedisServer):
        return obj.display.cpu()

    cpu_utilization.short_description = _('CPU Utilization')

    def slowlog(self, obj: RedisServer):
        return obj.display.slowlog()

    slowlog.short_description = _('Slowlog')

    def status(self, obj: RedisServer):
        return obj.stats.status

    status.short_description = _('Status')

    def memory(self, obj: RedisServer):
        return obj.stats.memory

    memory.short_description = _('Memory')

    def clients(self, obj: RedisServer):
        return obj.stats.clients

    clients.short_description = _('Clients')

    def tools(self, obj: RedisServer):
        return format_html(
            '<a href="{}">{}</a><br><a href="{}">{}</a>',
            resolve_url('admin:redisboard_redisserver_inspect', server_id=obj.id),
            _('Inspect'),
            resolve_url('admin:redisboard_redisserver_details', server_id=obj.id),
            _('Details'),
        )

    tools.short_description = _('Tools')

    def changelist_view(self, request, extra_context=None):
        response = super().changelist_view(request, extra_context)
        if isinstance(response, TemplateResponse):
            response.add_post_render_callback(cleanup_changelist_response)
        return response

    def get_urls(self):
        urlpatterns = super(RedisServerAdmin, self).get_urls()
        from django.urls import path

        def wrap(view):
            @wraps(view)
            @self.admin_site.admin_view
            def wrapper(request, server_id, **kwargs):
                server = get_object_or_404(RedisServer, id=server_id)
                if self.has_view_permission(request, server) and request.user.has_perm('redisboard.can_inspect'):
                    try:
                        response = view(request, server, **kwargs)
                        if isinstance(response, TemplateResponse):
                            response.add_post_render_callback(partial(cleanup_connection, server=server))
                        return response
                    finally:
                        cleanup_connection(None, server)
                else:
                    return HttpResponseForbidden("You can't inspect this server.")

            return wrapper

        return [
            path(
                '<int:server_id>/details/',
                wrap(self.details_view),
                name='redisboard_redisserver_details',
            ),
            path(
                '<int:server_id>/inspect/',
                wrap(self.inspect_view),
                name='redisboard_redisserver_inspect',
            ),
            path(
                '<int:server_id>/inspect/<int:db>/',
                wrap(self.inspect_view),
                name='redisboard_redisserver_inspect',
            ),
            path(
                '<int:server_id>/inspect/<int:db>/key/<path:key>/',
                wrap(self.inspect_key_view),
                name='redisboard_redisserver_inspect',
            ),
            path(
                '<int:server_id>/inspect/<int:db>/<int:cursor>/',
                wrap(self.inspect_view),
                name='redisboard_redisserver_inspect',
            ),
            path(
                '<int:server_id>/inspect/<int:db>/<int:cursor>/key/<path:key>/',
                wrap(self.inspect_key_view),
                name='redisboard_redisserver_inspect',
            ),
            path(
                '<int:server_id>/inspect/<int:db>/<int:cursor>/<int:count>/key/<path:key>/',
                wrap(self.inspect_key_view),
                name='redisboard_redisserver_inspect',
            ),
        ] + urlpatterns

    def inspect_key_view(self, request, server: RedisServer, db: int, key: str, cursor: int = 0, count: int = 0):
        key: bytes = unquote_to_bytes(key)
        display = server.display
        if not server.connection.exists(key):
            return HttpResponseNotFound('Key is gone.')
        stats = display.keys(db, [key])
        scan = display.value(db, key, cursor=cursor, count=count)
        return render(
            request,
            'redisboard/inspect_key.html',
            {
                **self.admin_site.each_context(request),
                'key': display.decoder.key(key),
                'encoded_key': quote(key),
                'stats': stats,
                'count': scan.count + count,
                'scan': scan,
                'db': {
                    'id': db,
                    'cursor': cursor,
                },
                'original': server,
                'opts': {'app_label': 'redisboard', 'object_name': 'redisserver'},
                'media': self.media,
            },
        )

    def inspect_view(self, request, server: RedisServer, db: Union[int, None] = None, cursor: Union[int, None] = 0):
        stats = server.stats
        active = None
        databases = []
        if stats:
            total_keys = sum(details.get('keys', 0) for details in stats.databases.values())
            if db is not None:
                if db in stats.databases:
                    active = DBInfo(db, stats.databases[db], scan=True)
                else:
                    active = DBInfo(db, {'keys': 0}, scan=True)
                databases = [active]
            elif total_keys < REDISBOARD_SCAN_COUNT:
                databases = [DBInfo(*item, scan=True) for item in stats.databases.items()]
            else:
                databases = [DBInfo(*item) for item in stats.databases.items()]

            for dbinfo in databases:
                if dbinfo.scan:
                    dbinfo.scan = server.display.scan(dbinfo.id, cursor=cursor, **request.GET.dict())
                    dbinfo.cursor = cursor
        return render(
            request,
            'redisboard/inspect.html',
            {
                **self.admin_site.each_context(request),
                'databases': databases,
                'display': server.display,
                'original': server,
                'stats': stats,
                'active': active,
                'filters': f'?{request.GET.urlencode()}' if request.GET else '',
                'opts': {'app_label': 'redisboard', 'object_name': 'redisserver'},
                'media': self.media,
            },
        )

    def details_view(self, request, server: RedisServer):
        sections = {}
        try:
            conn = server.connection
            conn.ping()
            status = 'UP'
        except (redis.exceptions.ConnectionError, redis.exceptions.TimeoutError) as exc:
            status = f'DOWN: {exc}'
        except redis.exceptions.RedisError as exc:
            status = f'ERROR: {exc!r}'
        else:
            for section in [
                'server',
                'clients',
                'memory',
                'persistence',
                'stats',
                'replication',
                'cpu',
                'commandstats',
                'latencystats',
                'cluster',
                'modules',
                'keyspace',
                'modules',
                'errorstats',
            ]:
                sections[section] = conn.info(section)

        return render(
            request,
            'redisboard/details.html',
            {
                **self.admin_site.each_context(request),
                'original': server,
                'sections': sections,
                'status': status,
                'opts': {'app_label': 'redisboard', 'object_name': 'redisserver'},
                'media': self.media,
            },
        )


admin.site.register(RedisServer, RedisServerAdmin)



import os
import re
import time
from pathlib import Path
from typing import Union

import psutil as psutil
import pytest
import requests
from process_tests import TestProcess
from process_tests import dump_on_error
from process_tests import wait_for_strings
from redis import StrictRedis
from redis.client import Pipeline

from redisboard.admin import cleanup_changelist_response
from redisboard.models import RedisServer

TIMEOUT = int(os.getenv('TEST_TIMEOUT', 60))
TEST_DATA_PATH = Path(__file__).with_name('test-data')
REALLY_LONG_STRING = 'foobar' * 1000


def make_test_data(conn: Union[StrictRedis, Pipeline], size=500):
    conn.set('my:str', 'bar')
    conn.set('my:big-str', 'foobar' * size)
    conn.set('my:bad-str', bytes(range(255)))
    conn.set('my:random-str', os.urandom(size))
    conn.set(f'bad:{REALLY_LONG_STRING}', 'big-str')
    conn.set(b'bad:' + bytes(range(255)), 'bad-str')
    conn.set(b'bad:' + os.urandom(1000), 'random-str')
    conn.sadd('my:set', 'bar')
    conn.sadd('my:set', REALLY_LONG_STRING)
    conn.sadd('my:set', bytes(range(255)))
    for i in range(size):
        conn.sadd('my:set', f'item-{i}')
    conn.sadd('my:random-set', os.urandom(1000))
    conn.hset('my:hash', 'str', 'bar')
    conn.hset('my:hash', 'big-str', REALLY_LONG_STRING)
    conn.hset('my:hash', 'bad-str', bytes(range(255)))
    for i in range(size):
        conn.hset('my:hash', f'item-{i}', 'bar')
    conn.hset('my:random-hash', 'random-str', os.urandom(1000))
    for i in range(size):
        conn.lpush('my:list', f'item-{i}')
    conn.zadd('my:zset', {f'item-{i}': i for i in range(size)})


@pytest.fixture
def redis_model(redis_server, db):
    server = RedisServer.objects.create(
        url=f'unix:///{redis_server}',
    )
    with server.connection:
        with server.connection.pipeline() as pipe:
            make_test_data(pipe)
            pipe.execute()
        server.connection.eval(
            """
            local c = 0
            while c < 1000000 do
                redis.call('get', 'str')
                c = c + 1
            end
        """,
            0,
        )
        yield server


@pytest.fixture
def empty_redis_model(redis_server, db):
    server = RedisServer.objects.create(
        url=f'unix:///{redis_server}?db=1',
    )
    with server.connection:
        yield server


@pytest.mark.django_db
def test_changelist(admin_client, redis_model):
    response = admin_client.get('/redisboard/redisserver/')
    assert cleanup_changelist_response in response._post_render_callbacks
    content = response.content.decode('utf-8')
    assert '>UP</' in content
    assert '>redis version</' in content
    assert '<a href="/redisboard/redisserver/{:d}/inspect/">Inspect</a>'.format(redis_model.pk) in content


@pytest.mark.django_db
def test_inspect(admin_client, redis_model):
    response = admin_client.get('/redisboard/redisserver/{:d}/inspect/'.format(redis_model.pk))
    content = response.content.decode('utf-8')
    assert 'Database 0:' in content
    assert '13 keys' in content
    assert '>Keys (13)</' in content
    assert '>UP</' in content
    assert '>redis version</' in content
    pk = redis_model.pk
    assert (
        f'<a href="/redisboard/redisserver/{pk}/inspect/0/key/bad%253A%2500%2501%2502%2503%2504%2505%2506%2507%2508%2509%250A%250B%250C'
        '%250D%250E%250F%2510%2511%2512%2513%2514%2515%2516%2517%2518%2519%251A%251B%251C%251D%251E%251F%2520%2521%2522%2523%2524%2525'
        '%2526%2527%2528%2529%252A%252B%252C-./0123456789%253A%253B%253C%253D%253E%253F%2540ABCDEFGHIJKLMNOPQRSTUVWXYZ%255B%255C%255D%'
        '255E_%2560abcdefghijklmnopqrstuvwxyz%257B%257C%257D~%257F%2580%2581%2582%2583%2584%2585%2586%2587%2588%2589%258A%258B%258C%25'
        '8D%258E%258F%2590%2591%2592%2593%2594%2595%2596%2597%2598%2599%259A%259B%259C%259D%259E%259F%25A0%25A1%25A2%25A3%25A4%25A5%25'
        'A6%25A7%25A8%25A9%25AA%25AB%25AC%25AD%25AE%25AF%25B0%25B1%25B2%25B3%25B4%25B5%25B6%25B7%25B8%25B9%25BA%25BB%25BC%25BD%25BE%25'
        'BF%25C0%25C1%25C2%25C3%25C4%25C5%25C6%25C7%25C8%25C9%25CA%25CB%25CC%25CD%25CE%25CF%25D0%25D1%25D2%25D3%25D4%25D5%25D6%25D7%25'
        'D8%25D9%25DA%25DB%25DC%25DD%25DE%25DF%25E0%25E1%25E2%25E3%25E4%25E5%25E6%25E7%25E8%25E9%25EA%25EB%25EC%25ED%25EE%25EF%25F0%25'
        'F1%25F2%25F3%25F4%25F5%25F6%25F7%25F8%25F9%25FA%25FB%25FC%25FD%25FE/">bad:\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r'
        '\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f !&quot;#$%&amp;&#x27;()*+,-./0123456789:;&lt;=&gt;?@'
        'ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\x7f\\x80\\x81\\x82\\x83\\x84\\x85\\x86\\x87\\x88\\x89\\x8a\\x'
        '8b\\x8c\\x8d\\x8e\\x8f\\x9…</a>' in content
    )
    assert f'<a href="/redisboard/redisserver/{pk}/inspect/0/key/my%253Ahash/">my:hash</a></td>' in content
    assert f'<a href="/redisboard/redisserver/{pk}/inspect/0/key/my%253Abig-str/">my:big-str</a></td>' in content
    assert f'<a href="/redisboard/redisserver/{pk}/inspect/0/key/my%253Arandom-str/">my:random-str</a></td>' in content
    assert f'<a href="/redisboard/redisserver/{pk}/inspect/0/key/my%253Azset/">my:zset</a></td>' in content
    assert f'<a href="/redisboard/redisserver/{pk}/inspect/0/key/my%253Aset/">my:set</a></td>' in content
    assert (
        f'<a href="/redisboard/redisserver/{pk}/inspect/0/key/bad%253A{REALLY_LONG_STRING}/">bad:{REALLY_LONG_STRING:.195}…</a>' in content
    )
    assert f'<a href="/redisboard/redisserver/{pk}/inspect/0/key/my%253Astr/">my:str</a></td>' in content
    assert f'<a href="/redisboard/redisserver/{pk}/inspect/0/key/my%253Alist/">my:list</a></td>' in content
    assert f'<a href="/redisboard/redisserver/{pk}/inspect/0/key/my%253Abad-str/">my:bad-str</a></td>' in content


@pytest.mark.django_db
def test_inspect_empty(admin_client, empty_redis_model):
    response = admin_client.get(f'/redisboard/redisserver/{empty_redis_model.pk}/inspect/')
    content = response.content.decode('utf-8')
    assert 'No database found' in content
    assert '>UP</' in content
    assert '>redis version</' in content


@pytest.mark.django_db
@pytest.mark.parametrize(
    'key',
    [
        'bad%253A%2500%2501%2502%2503%2504%2505%2506%2507%2508%2509%250A%2'
        '50B%250C%250D%250E%250F%2510%2511%2512%2513%2514%2515%2516%2517%2518%2519%251A%251B%251C%251D%251E%251F%2520%2521%2522%2523%25'
        '24%2525%2526%2527%2528%2529%252A%252B%252C-./0123456789%253A%253B%253C%253D%253E%253F%2540ABCDEFGHIJKLMNOPQRSTUVWXYZ%255B%255C'
        '%255D%255E_%2560abcdefghijklmnopqrstuvwxyz%257B%257C%257D~%257F%2580%2581%2582%2583%2584%2585%2586%2587%2588%2589%258A%258B%25'
        '8C%258D%258E%258F%2590%2591%2592%2593%2594%2595%2596%2597%2598%2599%259A%259B%259C%259D%259E%259F%25A0%25A1%25A2%25A3%25A4%25A'
        '5%25A6%25A7%25A8%25A9%25AA%25AB%25AC%25AD%25AE%25AF%25B0%25B1%25B2%25B3%25B4%25B5%25B6%25B7%25B8%25B9%25BA%25BB%25BC%25BD%25BE'
        '%25BF%25C0%25C1%25C2%25C3%25C4%25C5%25C6%25C7%25C8%25C9%25CA%25CB%25CC%25CD%25CE%25CF%25D0%25D1%25D2%25D3%25D4%25D5%25D6%25D7%'
        '25D8%25D9%25DA%25DB%25DC%25DD%25DE%25DF%25E0%25E1%25E2%25E3%25E4%25E5%25E6%25E7%25E8%25E9%25EA%25EB%25EC%25ED%25EE%25EF%25F0%2'
        '5F1%25F2%25F3%25F4%25F5%25F6%25F7%25F8%25F9%25FA%25FB%25FC%25FD%25FE',
        'my%253Ahash',
        'my%253Abig-str',
        'my%253Azset',
        'my%253Aset',
    ],
    ids='{:.16}'.format,
)
def test_key_details(admin_client, redis_model, key: str, request):
    response = admin_client.get(f'/redisboard/redisserver/{redis_model.pk}/inspect/0/key/{key}/')
    content = response.content.replace(b'\r', b'').replace(b'\n', b'')
    expected_path: Path = TEST_DATA_PATH.joinpath(f'{key.replace("%", ""):.16}.html')
    if expected_path.exists():
        expected = expected_path.read_bytes()
    else:
        expected = b'failfailfail'
    try:
        assert expected in content
    except Exception:
        if b'<fieldset class="module aligned key-data">' in content:
            content = content[content.index(b'<fieldset class="module aligned key-data">') : content.index(b'</fieldset></div>')]
        expected_path.write_bytes(content)
        raise


@pytest.mark.parametrize('entrypoint', ['redisboard', 'python -mredisboard'])
def test_cli(entrypoint, tmpdir):
    args = ['127.0.0.1:0', '--password', 'foobar', '--storage', str(tmpdir)]
    with TestProcess(*entrypoint.split() + args) as process:
        with dump_on_error(process.read):
            wait_for_strings(process.read, TIMEOUT, 'server at http://')
            t = time.time()
            port = None
            while time.time() - t < TIMEOUT and port is None:
                for conn in psutil.Process(process.proc.pid).connections('all'):
                    if conn.status == psutil.CONN_LISTEN and conn.laddr[0] == '127.0.0.1':
                        port = conn.laddr[1]
                        break
            assert port
            with requests.Session() as session:
                resp = session.get('http://127.0.0.1:%s/' % port)
                (csrftoken,) = re.findall('name=[\'"]csrfmiddlewaretoken[\'"] value=[\'"](.*?)[\'"]', resp.text)
                resp = session.post(
                    'http://127.0.0.1:%s/login/?next=/redisboard/redisserver/' % port,
                    data={
                        'csrfmiddlewaretoken': csrftoken,
                        'username': 'redisboard',
                        'password': 'foobar',
                    },
                )
                assert '<a href="/redisboard/redisserver/1/inspect/"' in resp.text
